@@ -24,14 +24,22 @@ use std::{fmt, io};
 
 use crate::consensus::encode::{self, serialize, Decodable, Encodable, VarInt};
 use crate::cryptonote::hash;
+use crate::cryptonote::onetime_key::KeyGenerator;
+use crate::util::key::H;
+use crate::{PublicKey, ViewPair};
 
 #[cfg(feature = "serde_support")]
 use serde::{Deserialize, Serialize};
 #[cfg(feature = "serde_support")]
 use serde_big_array_unchecked_docs::*;
 
+use curve25519_dalek::constants::ED25519_BASEPOINT_POINT;
+use curve25519_dalek::edwards::EdwardsPoint;
+use curve25519_dalek::scalar::Scalar;
 use thiserror::Error;
 
+use std::convert::TryInto;
+
 /// Serde support for array's bigger than 32 items.
 #[allow(missing_docs)]
 #[cfg(feature = "serde_support")]
@@ -44,8 +52,8 @@ pub mod serde_big_array_unchecked_docs {
 /// Ring Confidential Transaction potential errors.
 #[derive(Error, Debug, PartialEq)]
 pub enum Error {
-    /// Invalid RingCT type.
-    #[error("Unknown RingCT type")]
+    /// Invalid RingCt type.
+    #[error("Unknown RingCt type")]
     UnknownRctType,
 }
 
@@ -110,7 +118,7 @@ impl_consensus_encoding!(CtKey, mask);
 /// Multisig.
 #[derive(Debug)]
 #[allow(non_snake_case)]
-pub struct MultisigKLRki {
+pub struct MultisigKlrki {
     /// K value.
     pub K: Key,
     /// L value.
@@ -121,7 +129,7 @@ pub struct MultisigKLRki {
     pub ki: Key,
 }
 
-impl_consensus_encoding!(MultisigKLRki, K, L, R, ki);
+impl_consensus_encoding!(MultisigKlrki, K, L, R, ki);
 
 // ====================================================================
 /// Vector of multisig output keys.
@@ -153,6 +161,96 @@ pub enum EcdhInfo {
     },
 }
 
+impl EcdhInfo {
+    /// Opens the commitment and verifies it against the given one.
+    pub fn open_commitment(
+        &self,
+        view_pair: &ViewPair,
+        tx_pubkey: &PublicKey,
+        index: usize,
+        candidate_commitment: &EdwardsPoint,
+    ) -> Option<Opening> {
+        let shared_key = KeyGenerator::from_key(view_pair, *tx_pubkey).get_rvn_scalar(index);
+
+        let (amount, blinding_factor) = match self {
+            // ecdhDecode in rctOps.cpp else
+            EcdhInfo::Standard { mask, amount } => {
+                let shared_sec1 = hash::Hash::hash(shared_key.as_bytes()).to_bytes();
+                let shared_sec2 = hash::Hash::hash(&shared_sec1).to_bytes();
+                let mask_scalar = Scalar::from_bytes_mod_order(mask.key)
+                    - Scalar::from_bytes_mod_order(shared_sec1);
+
+                let amount_scalar = Scalar::from_bytes_mod_order(amount.key)
+                    - Scalar::from_bytes_mod_order(shared_sec2);
+                // get first 64 bits (d2b in rctTypes.cpp)
+                let amount_significant_bytes = amount_scalar.to_bytes()[0..8]
+                    .try_into()
+                    .expect("Can't fail");
+                let amount = u64::from_le_bytes(amount_significant_bytes);
+                (amount, mask_scalar)
+            }
+            // ecdhDecode in rctOps.cpp if (v2)
+            EcdhInfo::Bulletproof { amount } => {
+                let amount = xor_amount(amount.0, shared_key.scalar);
+                let mask = mask(shared_key.scalar);
+
+                (u64::from_le_bytes(amount), mask)
+            }
+        };
+
+        let amount_scalar = Scalar::from(amount);
+
+        let expected_commitment = ED25519_BASEPOINT_POINT * blinding_factor
+            + H.point.decompress().unwrap() * amount_scalar;
+
+        if &expected_commitment != candidate_commitment {
+            return None;
+        }
+
+        Some(Opening {
+            amount,
+            blinding_factor,
+            commitment: expected_commitment,
+        })
+    }
+}
+
+/// The result of opening the commitment inside the transaction.
+#[derive(Debug)]
+pub struct Opening {
+    /// The original amount of the output.
+    pub amount: u64,
+    /// The blinding factor used to blind the amount.
+    pub blinding_factor: Scalar,
+    /// The commitment used to verify the blinded amount.
+    pub commitment: EdwardsPoint,
+}
+
+fn xor_amount(amount: [u8; 8], shared_key: Scalar) -> [u8; 8] {
+    // ecdhHash in .cpp
+    let mut amount_key = b"amount".to_vec();
+    amount_key.extend(shared_key.as_bytes());
+
+    // Hn("amount", Hn(rKbv,t))
+    let hash_shared_key = hash::Hash::hash(&amount_key).to_fixed_bytes();
+    let hash_shared_key_significant_bytes = hash_shared_key[0..8]
+        .try_into()
+        .expect("hash_shared_key create above has 32 bytes");
+
+    // amount_t = bt XOR Hn("amount", Hn("amount", Hn(rKbv,t)))
+    // xor8(masked.amount, ecdhHash(sharedSec)); in .cpp
+    (u64::from_le_bytes(amount) ^ u64::from_le_bytes(hash_shared_key_significant_bytes))
+        .to_le_bytes()
+}
+
+fn mask(scalar: Scalar) -> Scalar {
+    let mut commitment_key = b"commitment_mask".to_vec();
+    commitment_key.extend(scalar.as_bytes());
+
+    // yt in Z2M p 53
+    hash::Hash::hash_to_scalar(&commitment_key).scalar
+}
+
 impl fmt::Display for EcdhInfo {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
         match self {
@@ -171,7 +269,7 @@ impl fmt::Display for EcdhInfo {
 }
 
 impl EcdhInfo {
-    /// Decode Diffie-Hellman info given the RingCT type.
+    /// Decode Diffie-Hellman info given the RingCt type.
     fn consensus_decode<D: io::Read>(
         d: &mut D,
         rct_type: RctType,
@@ -183,7 +281,7 @@ impl EcdhInfo {
                     amount: Decodable::consensus_decode(d)?,
                 })
             }
-            RctType::Bulletproof2 | RctType::CLSAG => Ok(EcdhInfo::Bulletproof {
+            RctType::Bulletproof2 | RctType::Clsag => Ok(EcdhInfo::Bulletproof {
                 amount: Decodable::consensus_decode(d)?,
             }),
         }
@@ -222,7 +320,7 @@ pub struct BoroSig {
 impl_consensus_encoding!(BoroSig, s0, s1, ee);
 
 // ====================================================================
-/// Contains the necessary keys to represent MLSAG signature.
+/// Contains the necessary keys to represent Mlsag signature.
 #[derive(Debug, Clone)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
 pub struct MgSig {
@@ -244,11 +342,11 @@ impl Encodable for MgSig {
 }
 
 // ====================================================================
-/// CLSAG signature.
+/// Clsag signature.
 #[derive(Debug, Clone)]
 #[allow(non_snake_case)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
-pub struct CLSAG {
+pub struct Clsag {
     /// scalars.
     pub s: Vec<Key>,
     /// c1 value.
@@ -257,7 +355,7 @@ pub struct CLSAG {
     pub D: Key,
 }
 
-impl Encodable for CLSAG {
+impl Encodable for Clsag {
     fn consensus_encode<S: io::Write>(&self, s: &mut S) -> Result<usize, io::Error> {
         let mut len = 0;
         // Encode the vector without prefix lenght
@@ -315,11 +413,11 @@ pub struct Bulletproof {
 impl_consensus_encoding!(Bulletproof, A, S, T1, T2, taux, mu, L, R, a, b, t);
 
 // ====================================================================
-/// RingCT base signature format.
+/// RingCt base signature format.
 #[derive(Debug, Clone)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
 pub struct RctSigBase {
-    /// The RingCT type of signatures.
+    /// The RingCt type of signatures.
     pub rct_type: RctType,
     /// Transaction fee.
     pub txn_fee: VarInt,
@@ -349,7 +447,7 @@ impl fmt::Display for RctSigBase {
 }
 
 impl RctSigBase {
-    /// Decode a RingCT base signature given the number of inputs and outputs of the transaction.
+    /// Decode a RingCt base signature given the number of inputs and outputs of the transaction.
     pub fn consensus_decode<D: io::Read>(
         d: &mut D,
         inputs: usize,
@@ -368,7 +466,7 @@ impl RctSigBase {
             | RctType::Simple
             | RctType::Bulletproof
             | RctType::Bulletproof2
-            | RctType::CLSAG => {
+            | RctType::Clsag => {
                 let mut pseudo_outs: Vec<Key> = vec![];
                 // TxnFee
                 let txn_fee: VarInt = Decodable::consensus_decode(d)?;
@@ -405,7 +503,7 @@ impl Encodable for RctSigBase {
             | RctType::Simple
             | RctType::Bulletproof
             | RctType::Bulletproof2
-            | RctType::CLSAG => {
+            | RctType::Clsag => {
                 len += self.txn_fee.consensus_encode(s)?;
                 if self.rct_type == RctType::Simple {
                     len += encode_sized_vec!(self.pseudo_outs, s);
@@ -439,8 +537,8 @@ pub enum RctType {
     Bulletproof,
     /// Bulletproof2 type.
     Bulletproof2,
-    /// CLSAG Ring signatures, used in the current network.
-    CLSAG,
+    /// Clsag Ring signatures, used in the current network.
+    Clsag,
 }
 
 impl fmt::Display for RctType {
@@ -451,7 +549,7 @@ impl fmt::Display for RctType {
             RctType::Simple => "Simple",
             RctType::Bulletproof => "Bulletproof",
             RctType::Bulletproof2 => "Bulletproof2",
-            RctType::CLSAG => "CLSAG",
+            RctType::Clsag => "Clsag",
         };
         write!(fmt, "{}", rct_type)
     }
@@ -462,7 +560,7 @@ impl RctType {
     pub fn is_rct_bp(self) -> bool {
         matches!(
             self,
-            RctType::Bulletproof | RctType::Bulletproof2 | RctType::CLSAG
+            RctType::Bulletproof | RctType::Bulletproof2 | RctType::Clsag
         )
     }
 }
@@ -476,7 +574,7 @@ impl Decodable for RctType {
             2 => Ok(RctType::Simple),
             3 => Ok(RctType::Bulletproof),
             4 => Ok(RctType::Bulletproof2),
-            5 => Ok(RctType::CLSAG),
+            5 => Ok(RctType::Clsag),
             _ => Err(Error::UnknownRctType.into()),
         }
     }
@@ -490,13 +588,13 @@ impl Encodable for RctType {
             RctType::Simple => 2u8.consensus_encode(s),
             RctType::Bulletproof => 3u8.consensus_encode(s),
             RctType::Bulletproof2 => 4u8.consensus_encode(s),
-            RctType::CLSAG => 5u8.consensus_encode(s),
+            RctType::Clsag => 5u8.consensus_encode(s),
         }
     }
 }
 
 // ====================================================================
-/// Prunable part of RingCT signature format.
+/// Prunable part of RingCt signature format.
 #[derive(Debug, Clone)]
 #[allow(non_snake_case)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
@@ -508,14 +606,14 @@ pub struct RctSigPrunable {
     /// MSLAG signatures, simple rct has N, full has 1.
     pub MGs: Vec<MgSig>,
     /// CSLAG signatures.
-    pub CLSAGs: Vec<CLSAG>,
+    pub Clsags: Vec<Clsag>,
     /// Pseudo out vector.
     pub pseudo_outs: Vec<Key>,
 }
 
 impl RctSigPrunable {
-    /// Decode a prunable RingCT signature given the number of inputs and outputs in the
-    /// transaction, the RingCT type and the number of mixins.
+    /// Decode a prunable RingCt signature given the number of inputs and outputs in the
+    /// transaction, the RingCt type and the number of mixins.
     #[allow(non_snake_case)]
     pub fn consensus_decode<D: io::Read>(
         d: &mut D,
@@ -530,12 +628,12 @@ impl RctSigPrunable {
             | RctType::Simple
             | RctType::Bulletproof
             | RctType::Bulletproof2
-            | RctType::CLSAG => {
+            | RctType::Clsag => {
                 let mut bulletproofs: Vec<Bulletproof> = vec![];
                 let mut range_sigs: Vec<RangeSig> = vec![];
                 if rct_type.is_rct_bp() {
                     match rct_type {
-                        RctType::Bulletproof2 | RctType::CLSAG => {
+                        RctType::Bulletproof2 | RctType::Clsag => {
                             bulletproofs = Decodable::consensus_decode(d)?;
                         }
                         _ => {
@@ -547,11 +645,11 @@ impl RctSigPrunable {
                     range_sigs = decode_sized_vec!(outputs, d);
                 }
 
-                let mut CLSAGs: Vec<CLSAG> = vec![];
+                let mut Clsags: Vec<Clsag> = vec![];
                 let mut MGs: Vec<MgSig> = vec![];
 
                 match rct_type {
-                    RctType::CLSAG => {
+                    RctType::Clsag => {
                         for _ in 0..inputs {
                             let mut s: Vec<Key> = vec![];
                             for _ in 0..=mixin {
@@ -560,7 +658,7 @@ impl RctSigPrunable {
                             }
                             let c1 = Decodable::consensus_decode(d)?;
                             let D = Decodable::consensus_decode(d)?;
-                            CLSAGs.push(CLSAG { s, c1, D });
+                            Clsags.push(Clsag { s, c1, D });
                         }
                     }
                     _ => {
@@ -583,7 +681,7 @@ impl RctSigPrunable {
 
                 let mut pseudo_outs: Vec<Key> = vec![];
                 match rct_type {
-                    RctType::Bulletproof | RctType::Bulletproof2 | RctType::CLSAG => {
+                    RctType::Bulletproof | RctType::Bulletproof2 | RctType::Clsag => {
                         pseudo_outs = decode_sized_vec!(inputs, d);
                     }
                     _ => (),
@@ -593,14 +691,14 @@ impl RctSigPrunable {
                     range_sigs,
                     bulletproofs,
                     MGs,
-                    CLSAGs,
+                    Clsags,
                     pseudo_outs,
                 }))
             }
         }
     }
 
-    /// Encode the prunable RingCT signature part given the RingCT type of the transaction.
+    /// Encode the prunable RingCt signature part given the RingCt type of the transaction.
     pub fn consensus_encode<S: io::Write>(
         &self,
         s: &mut S,
@@ -612,11 +710,11 @@ impl RctSigPrunable {
             | RctType::Simple
             | RctType::Bulletproof
             | RctType::Bulletproof2
-            | RctType::CLSAG => {
+            | RctType::Clsag => {
                 let mut len = 0;
                 if rct_type.is_rct_bp() {
                     match rct_type {
-                        RctType::Bulletproof2 | RctType::CLSAG => {
+                        RctType::Bulletproof2 | RctType::Clsag => {
                             len += self.bulletproofs.consensus_encode(s)?;
                         }
                         _ => {
@@ -630,12 +728,12 @@ impl RctSigPrunable {
                 }
 
                 match rct_type {
-                    RctType::CLSAG => len += encode_sized_vec!(self.CLSAGs, s),
+                    RctType::Clsag => len += encode_sized_vec!(self.Clsags, s),
                     _ => len += encode_sized_vec!(self.MGs, s),
                 }
 
                 match rct_type {
-                    RctType::Bulletproof | RctType::Bulletproof2 | RctType::CLSAG => {
+                    RctType::Bulletproof | RctType::Bulletproof2 | RctType::Clsag => {
                         len += encode_sized_vec!(self.pseudo_outs, s);
                     }
                     _ => (),
@@ -647,7 +745,7 @@ impl RctSigPrunable {
 }
 
 // ====================================================================
-/// A RingCT signature.
+/// A RingCt signature.
 #[derive(Debug, Clone, Default)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
 pub struct RctSig {
